/*
  hppNextBestViewPlanner.h.cpp -

  Copyright 2008 Centre National de la Recherche Scientifique (CNRS)

  Authors: Alireza Nakhaei,

*/
/*****************************************
 INCLUDES
*******************************************/

#include <iostream>
#include "hppWorldModelPlanner/hppWorldModelDelegate.h"
#include "hppWorldModelPlanner/hppWorldModelPlanner.h"
#include "hppWorldModelPlanner/kwsPlusShooter.h"
#include "hppWalkFootPlanner/hppWalkDistance.h"

#include "KineoWorks2/kwsShooterPath.h"
#include "kcd2/kcdAnalysis.h"
#include "KineoWorks2/kwsConfig.h"
#include "KineoWorks2/kwsNode.h"

#include "KineoModel/kppDofComponent.h"

#include "kwsPlus/hppVisRdmBuilder.h"
#include "kwsPlus/hppShooterActiveDof.h"

#define OPT_MAX_NB	(50)

//-------------------------------------------------------------------------
// Select verbosity at configuration by setting CXXFLAGS="... -DDEBUG=[1 or 2]"
#if DEBUG==2
#define ODEBUG2(x) std::cout << "hppWorldModelPlanner:" << x << std::endl
#define ODEBUG1(x) std::cerr << "hppWorldModelPlanner:" << x << std::endl
#elif DEBUG==1
#define ODEBUG2(x)
#define ODEBUG1(x) std::cerr << "hppWorldModelPlanner:" << x << std::endl
#else
#define ODEBUG2(x)
#define ODEBUG1(x)
#endif
//---------------------------------------------------------------------------

ChppWorldModelPlanner::ChppWorldModelPlanner ( double inStepLength, double inStepSpace, double inSecurityDistance,
        double inBoxLateralTolerance, const char *inModel ) :ChppWalkPlanner ( inStepLength, inStepSpace, inSecurityDistance, inBoxLateralTolerance, inModel )
{
	attNextBestViewPlanner = new ChppNextBestViewPlanner();
	Tps = ( struct timeval* ) malloc ( sizeof ( struct timeval ) );
	Tpf = ( struct timeval* ) malloc ( sizeof ( struct timeval ) );
	Tzp = 0;
	ODEBUG2 ( "Constructor of class ChppWorldModelPlanner." );
}
// ==========================================================================
ChppWorldModelPlanner::~ChppWorldModelPlanner()
{
	free ( Tps );
	free ( Tpf );
	ODEBUG2 ( "------------------------------------------------------------------" );
	ODEBUG2 ( "              Destructor of class ChppWorldModelPlanner." );
	ODEBUG2 ( "------------------------------------------------------------------" );
}
// ==========================================================================
ktStatus ChppWorldModelPlanner::setResolution ( double inResolution )
{
	ODEBUG2 ( "attSlamWalkPlanner:setResolution ..." );

	attOccupiedEnv = ChppPolyhedronMap::create ( "occupied" , inResolution );
	attOccupiedEnv->diffuseColor ( CkppColor ( 1.0f, 0.0f, 0.0f, 1.0f ) );
	attUnknownEnv = ChppPolyhedronMap::create ( "unknown" , inResolution );
	attUnknownEnv->diffuseColor ( CkppColor ( 0.0f, 1.0f, 0.0f, 1.0f ) );
	attNextBestViewPlanner->setResolution ( inResolution );
	attNextBestViewPlanner->setCameraParameters ( 1, 1, 2 );
	attNextBestViewPlanner->setCoveringAngle ( -1.1,.2,-.7,.7 );
	ODEBUG2 ( "attSlamWalkPlanner:setResolution end." );

	return KD_OK;
}
// ==========================================================================
ktStatus ChppWorldModelPlanner::allocateMap ( double xstart, double ystart, double zstart,
        double xend, double yend, double zend )
{
	ODEBUG2 ( "attSlamWalkPlanner:allocateMap ..." );
	att3DMapArea.xstart=xstart;
	att3DMapArea.ystart=ystart;
	att3DMapArea.zstart=zstart;
	att3DMapArea.xend=xend;
	att3DMapArea.yend=yend;
	att3DMapArea.zend=zend;
	ODEBUG2 ( "attSlamWalkPlanner:allocateMap end." );

	return KD_OK;
}
// ==========================================================================
ktStatus ChppWorldModelPlanner::initializeProblem()
{
	ODEBUG2 ( " : initializeProblem ..." );
	srand ( time ( NULL ) );

	if ( ChppWalkPlanner::initializeProblem() !=KD_OK )
	{
		ODEBUG1 ( " : initializeProblem .: initialize hppWalkPlanner failed." );
		return KD_ERROR;
	}

	CkppDeviceComponentShPtr robotBox =  robotIthProblem ( 0 );
	ODEBUG2 ( " : initializeProblem : : Device Name : " << robotBox->name() );
	if ( !robotBox )
	{
		ODEBUG1 ( ":initializeProblem: failed to get the robotBox." );
		return KD_ERROR;
	}
	attNextBestViewPlanner->initializeProblem ( robotBox );

	//-------------------------------create RoadMap
	CkwsPlusRoadmapShPtr kwsPlusRoadmap = CkwsPlusRoadmap::create ( robotBox ) ;
	if ( !kwsPlusRoadmap )
	{
		ODEBUG1 ( ":initializeProblem: ERROR : No kwsPlusRoadmap" );
		return KD_ERROR;
	}

	//ChppVisRdmBuilderShPtr kwsRoadmapBuilder = ChppVisRdmBuilder::create ( kwsPlusRoadmap, 0.05 );
	//ChppVisRdmBuilderShPtr kwsRoadmapBuilder = ChppVisRdmBuilder::create ( kwsPlusRoadmap, 0.1, shooterCfgSp, distance_evaluator );
	//CkwsDiffusingRdmBuilderShPtr kwsRoadmapBuilder = CkwsDiffusingRdmBuilder::create ( kwsPlusRoadmap, 0.1,distance_evaluator,picker, shooterCfgSp );
	CkwsDiffusingRdmBuilderShPtr kwsRoadmapBuilder = CkwsDiffusingRdmBuilder::create ( kwsPlusRoadmap, 0.05 );
	if ( !kwsRoadmapBuilder )
	{
		ODEBUG1 ( ": initializeProblem: ERROR : No kwsRoadmapBuilder" );
		return KD_ERROR;
	}

	//kwsRoadmapBuilder->diffuseFromProblemStart ( true );
	//kwsRoadmapBuilder->diffuseFromProblemGoal ( true );

	kwsPlusRoadmap->setRoadmapBuilder ( kwsRoadmapBuilder );

	ChppWorldModelDelegate * roadmapDelegate= new ChppWorldModelDelegate ( 150 );
	if ( !roadmapDelegate )
	{
		ODEBUG1 ( ": initializeProblem:ERROR - No Delegate" );
		return KD_ERROR;
	}

	if ( kwsRoadmapBuilder->addDelegate ( roadmapDelegate ) != KD_OK )
	{
		ODEBUG1 ( ": initializeProblem:Error : addDelegate Failed." );
		return KD_ERROR;
	}

	
	roadmapBuilderIthProblem ( 0, kwsRoadmapBuilder, true );

	// Create a Path Optimizer
	CkwsDistanceShPtr distance = robotBox->specificDistance();
	if ( !distance )
	{
		ODEBUG1 ( "hppWorldModelPlanner::Init: No distance." );
		return KD_ERROR;
	}
	CkwsLoopOptimizerShPtr pathOptimizer = CkwsRandomOptimizer::create();
	pathOptimizer->distance ( distance );
	pathOptimizer->shortcutMethod ( CkwsShortcutDirect::create() );
	pathOptimizer->maxNbLoop ( OPT_MAX_NB );

	pathOptimizerIthProblem ( 0, pathOptimizer );

//-----------  Setting the environmet size for hppWorldModelPlanner
	CkppJointComponentShPtr kppJoint = robotBox->rootJointComponent();
	if ( !kppJoint )
	{
		ODEBUG1 ( "hppWorldModelPlanner::Init: bounding box robot has no root joint(hppWorldModelPlanner)." );
		return KD_ERROR;
	}

	CkwsJointShPtr kwsJoint = kppJoint->kwsJoint();
	if ( kwsJoint->countDofs() != 6 )
	{
		ODEBUG1 ( "hppWorldModelPlanner::Init: Root joint of bounding box is not freeflyer(hppWorldModelPlanner)." );
		return KD_ERROR;
	}
	ODEBUG2 ( "TEST : Root joint of bounding box countDof = " << kwsJoint->countDofs() );

	kwsJoint->dof ( 0 )->isBounded ( true );
	kwsJoint->dof ( 0 )->bounds ( att3DMapArea.xstart, att3DMapArea.xend );
	kwsJoint->dof ( 1 )->isBounded ( true );
	kwsJoint->dof ( 1 )->bounds ( att3DMapArea.ystart, att3DMapArea.yend );
	kwsJoint->dof ( 2 )->isLocked ( true );
	kwsJoint->dof ( 3 )->isLocked ( true );
	kwsJoint->dof ( 4 )->isLocked ( true );
	kwsJoint->dof ( 5 )->isBounded ( true );
	kwsJoint->dof ( 5 )->bounds ( -2*M_PI, 2*M_PI );
	robotBox->extraDof ( 0 )->CkwsDof::bounds ( -2, 2 );

//------------------------------------------------------------ Create Dynamic Environment handler
	ChppDeviceShPtr hppDevice =  KIT_DYNAMIC_PTR_CAST ( ChppDevice, robotIthProblem ( 0 ) );
	if ( !hppDevice )
		ODEBUG1 ( "initializeProblem :ERROR - NO hppDevice" );

	attDynamicEnvHandler=ChppDynamicEnvHandler::create ( hppDevice );
	attDynamicEnvHandler->setCellSize ( 2 );
	attDynamicEnvHandler->allocateMap ( att3DMapArea.xstart, att3DMapArea.ystart, att3DMapArea.zstart, att3DMapArea.xend, att3DMapArea.yend, att3DMapArea.zend );
	attDynamicEnvHandler->setRoadmap ( kwsPlusRoadmap );

	attDynamicObstacle = ChppDynamicObstacleOccupancyMap::create ( attOccupiedEnv ,attDynamicEnvHandler );
	attDynamicUnknownEnv = ChppDynamicObstacleOccupancyMap::create ( attUnknownEnv ,attDynamicEnvHandler );

	attDynamicEnvHandler->addObstacle ( attDynamicObstacle );
	attDynamicEnvHandler->addObstacle ( attDynamicUnknownEnv );
	
//------------------------------------------------------------initialize next bext view planner
	ODEBUG2 ( "TEST : Robot box original" );
	ODEBUG2 ( "TEST :  NUMBER OF DOF = " << robotBox-> countDofs() );
	ODEBUG2 ( "TEST : NUMBER OF EXTRA DOF = " << robotBox-> countExtraDofs() );

	attNextBestViewPlanner->allocateMap ( att3DMapArea.xstart, att3DMapArea.ystart, att3DMapArea.zstart, att3DMapArea.xend, att3DMapArea.yend, att3DMapArea.zend );
	removeEnv();

	return KD_OK;
}
// ==========================================================================
ktStatus ChppWorldModelPlanner::solve()
{
	ODEBUG2 ( "solve ..." );
	attOccupiedEnv->getPolyhedronMap()-> tolerance (0.01);
	attUnknownEnv->getPolyhedronMap()-> tolerance (0.01);

	std::vector<double> vecConfig;
	CkwsConfigShPtr confTest ;
	confTest = initConfIthProblem ( 0 );
	confTest->getDofValues ( vecConfig );
	ODEBUG2 ( "INIT CONFIG in  Planner : " << vecConfig[1] << ", " << vecConfig[2] << ", " << vecConfig[6] );

	confTest = goalConfIthProblem ( 0 );
	confTest->getDofValues ( vecConfig );
	ODEBUG2 ( "GOAL CONFIG in  Planner: " << vecConfig[1] << ", " << vecConfig[2] << " , " << vecConfig[6] );

	gettimeofday ( Tps, Tzp );
	//CkcdAnalysis::resetComputeCount();

	if ( ChppWalkPlanner::solve() ==KD_OK )
	{
		gettimeofday ( Tpf, Tzp );
		ODEBUG2 ( "TIME: Solving Problem =: " << ( Tpf->tv_sec-Tps->tv_sec ) *1000 + ( Tpf->tv_usec-Tps->tv_usec ) /1000 << " ms" );
		//cout << "Compute count = " << CkcdAnalysis::computeCount() << endl;
		ODEBUG2 ( " Problem was solved successfully." );
	}
	else
	{
		ODEBUG2 ( "There is no solution." );
		return KD_ERROR;
	}
	attDynamicEnvHandler->resetEnvHandler();
	ODEBUG2 ( "solve end." );
	return KD_OK ;
}
//==================================================================================================
ktStatus ChppWorldModelPlanner::updateEnvironmentKppPolyhedron ( const std::vector<char> inRangeMap )
{
	ODEBUG2 ( " updateEnvironmentKppPolyhedron .... " );
	gettimeofday ( Tps, Tzp );

	std::vector<char> occupiedRangeMap;
	std::vector<char> unknownEnvRangeMap;

	ODEBUG2 ( "  updateEnvironmentKppPolyhedron: inRangeMap.size()= " << inRangeMap.size() );

	occupiedRangeMap.resize ( inRangeMap.size() );
	unknownEnvRangeMap.resize ( inRangeMap.size() );
	unsigned int nbOccupied=0;
	unsigned int nbUnknown=0;
	for ( unsigned int i=0 ; i< inRangeMap.size(); i++ )
	{
		if ( inRangeMap.at ( i ) ==1 )
		{
			occupiedRangeMap.at ( i ) =1;
			nbOccupied++;
		}
		else
			occupiedRangeMap.at ( i ) =0;

		if ( inRangeMap.at ( i ) ==2 )
		{
			unknownEnvRangeMap.at ( i ) =1;
			nbUnknown++;
		}
		else
			unknownEnvRangeMap.at ( i ) =0;
	}
	ODEBUG2 ( "  updateEnvironmentKppPolyhedron:  nbOccupied= " << nbOccupied << " , nbUnknown=  " << nbUnknown );

	if ( !attDynamicObstacle )
		ODEBUG2 ( "updateEnvironmentKppPolyhedron : ERROR NO attDynamicObstacle" );

	attDynamicObstacle->updateEnvironmentKppPolyhedron ( occupiedRangeMap );
	attDynamicUnknownEnv->updateEnvironmentKppPolyhedron ( unknownEnvRangeMap );

	gettimeofday ( Tpf, Tzp );
	ODEBUG2 ( "TIME: Update Environment =: " << ( Tpf->tv_sec-Tps->tv_sec ) *1000 + ( Tpf->tv_usec-Tps->tv_usec ) /1000 << " ms" );

	attNextBestViewPlanner->updateRangeMap ( inRangeMap );

	ODEBUG2 ( "updateEnvironmentKppPolyhedron end." );
	return KD_OK;
}
//====================================================================================================
ktStatus ChppWorldModelPlanner::loadOccupiedEnv()
{

	ODEBUG2 ( "loadOccupiedEnv ..." );
	std::vector<CkcdObjectShPtr> listOfObstacle;
	listOfObstacle.push_back ( attOccupiedEnv->getPolyhedronMap() );
	obstacleList ( listOfObstacle );
	attNextBestViewPlanner->obstacleList ( listOfObstacle );
	ODEBUG2 ( "loadOccupiedEnv end." );

	return KD_OK;
}

//====================================================================================================
ktStatus ChppWorldModelPlanner::loadCompleteEnv()
{
	ODEBUG2 ( "loadCompleteEnv ..." );
	std::vector<CkcdObjectShPtr> listOfObstacle;
	attOccupiedEnv->getPolyhedronMap()->tolerance (0.01);
	attUnknownEnv->getPolyhedronMap()->tolerance (0.01);
	
	listOfObstacle.push_back ( attOccupiedEnv->getPolyhedronMap() );
	listOfObstacle.push_back ( attUnknownEnv->getPolyhedronMap() );
	
	
	obstacleList ( listOfObstacle );
	attNextBestViewPlanner->obstacleList ( listOfObstacle );

	ODEBUG2 ( "loadCompleteEnv end." );

	return KD_OK;
}

//====================================================================================================
ktStatus ChppWorldModelPlanner::removeEnv()
{
	ODEBUG2 ( "removeEnv ..." );
	std::vector<CkcdObjectShPtr> listOfObstacle;
	obstacleList ( listOfObstacle );
	attNextBestViewPlanner->obstacleList ( listOfObstacle );

	ODEBUG2 ( "removeEnv end." );

	return KD_OK;
}
//====================================================================================================

ktStatus ChppWorldModelPlanner::updateRoadmap()
{
	ODEBUG2 ( "updateRoadmap..... " );
	gettimeofday ( Tps, Tzp );
	if ( clearPaths() !=KD_OK )
	{
		ODEBUG1 ( "Error in clearing path" );
		return KD_ERROR;
	}
	ODEBUG2 ( "Start of Updating Roadmap " );
	if (	attDynamicEnvHandler->updateRoadmap() !=KD_OK )
	{
		cout <<"Error in Updating Roadmap"<< endl;
		return KD_ERROR;
	}
	gettimeofday ( Tpf, Tzp );

	ODEBUG2 ( "TIME: For Updating Roadmap: " << ( Tpf->tv_sec-Tps->tv_sec ) *1000 + ( Tpf->tv_usec-Tps->tv_usec ) /1000 << " ms" ) ;

	ODEBUG2 ( "updateRoadmap end." );
	return KD_OK;
}

//===============================================================================================
ktStatus ChppWorldModelPlanner::clearPaths()
{
	ODEBUG2 ( "ClearPaths ..... " );

	int nbProblem = getNbHppProblems();
	ODEBUG2 ( "clearPaths-number of problem = " << nbProblem );

// Clearing path in problem
	for ( int i = 0; i< nbProblem; i++ )
	{
		unsigned int pathID=0;
		while ( ( pathID<getNbPaths ( i ) ) )
		{
			CkwsPathShPtr kwsPathWalk = getPath ( i , pathID );
			ODEBUG2 ( "clearPaths-clear() " );

			kwsPathWalk->clear();
			pathID++;
		}
	}

	ODEBUG2 ( "clearPaths-number of path in 2 problem = " << getNbPaths ( 0 ) << " , " << getNbPaths ( 1 ) );

	ODEBUG2 ( "ClearPaths ends. " );
	return KD_OK;
}
//===============================================================================================
ktStatus ChppWorldModelPlanner::clearRoadmap()
{
	ODEBUG2 ( "Clearing Roadmap ..... " );
	clearPaths();
	ODEBUG2 ( "All Paths were cleard " );
	CkwsRoadmapShPtr roadmap = ChppPlanner::roadmapBuilderIthProblem ( 0 )->roadmap();
	roadmap->clear();
	ODEBUG2 ( "......Roadmap was cleared " );
	return KD_OK;
}
//==========================================================================================================
std::vector<double> ChppWorldModelPlanner::nextBextView ( double initX, double initY, double initTheta,double goalX, double goalY, double goalTheta )
{
	return attNextBestViewPlanner->nextBextView ( initX, initY, initTheta,goalX, goalY, goalTheta );
}
//==========================================================================================================
std::vector<double> ChppWorldModelPlanner::nextBextView ( double initX, double initY, double initTheta)
{
	return attNextBestViewPlanner->nextBextView ( initX, initY, initTheta);
}
//==========================================================================================================
double ChppWorldModelPlanner::visibility ( double inX, double inY, double inTheta, double goalX, double goalY, double goalTheta )
{
	double vis = attNextBestViewPlanner->visibility ( inX, inY, inTheta, goalX, goalY, goalTheta );
	ODEBUG2 ( "Visibility is = " << vis );
	return vis;
}
//==========================================================================================================
double ChppWorldModelPlanner::visibility ( double inX, double inY, double inTheta )
{
	ODEBUG2 ( "Visibility  start ... ");
	double vis = attNextBestViewPlanner->visibility ( inX, inY, inTheta );
	ODEBUG2 ( "Visibility is = " << vis );
	return vis;
}
